#! /usr/bin/env python3
#
# Example program using irc.client.
#
# This program is free without restrictions; do anything you like with
# it.
# IMPORTANT: sasl_login must equal your nickserv account name
#
# Matthew Blau <mrb1105@gmail.com>

import functools
import ssl
import sys

import irc.client
import irc


class IRCCat(irc.client.SimpleIRCClient):
    def __init__(self, target):
        irc.client.SimpleIRCClient.__init__(self)
        self.target = target

    def on_welcome(self, connection, event):
        if irc.client.is_channel(self.target):
            connection.join(self.target)
        else:
            self.send_it()

    def on_login_failed(self, connection, event):
        print(event)

    def on_join(self, connection, event):
        self.send_it()

    def on_disconnect(self, connection, event):
        sys.exit(0)

    def send_it(self):
        while 1:
            line = sys.stdin.readline().strip()
            if not line:
                break
            self.connection.privmsg(self.target, line)
        self.connection.quit("Using irc.client.py")


def main():
    server = "irc.libera.chat"
    port = 6697
    nickname = "nickname"
    account_name = "username"
    target = "##channel"
    password = ""

    c = IRCCat(target)
    try:
        context = ssl.create_default_context()
        wrapper = functools.partial(context.wrap_socket, server_hostname=server)

        c.connect(
            server,
            port,
            nickname,
            password,
            sasl_login=account_name,
            username=account_name,
            connect_factory=irc.connection.Factory(wrapper=wrapper),
        )
    except irc.client.ServerConnectionError as x:
        print(x)
        sys.exit(1)
    c.start()


if __name__ == "__main__":
    main()



from unittest import mock

import pytest

import irc.client


def test_version():
    assert isinstance(irc._get_version(), str)


@mock.patch('irc.connection.socket')
def test_privmsg_sends_msg(socket_mod):
    server = irc.client.Reactor().server()
    server.connect('foo', 6667, 'bestnick')
    # make sure the mock object doesn't have a write method or it will treat
    #  it as an SSL connection and never call .send.
    del server.socket.write
    server.privmsg('#best-channel', 'You are great')
    server.socket.send.assert_called_with(b'PRIVMSG #best-channel :You are great\r\n')


@mock.patch('irc.connection.socket')
def test_privmsg_fails_on_embedded_carriage_returns(socket_mod):
    server = irc.client.Reactor().server()
    server.connect('foo', 6667, 'bestnick')
    with pytest.raises(ValueError):
        server.privmsg('#best-channel', 'You are great\nSo are you')


class TestHandlers:
    def test_handlers_same_priority(self):
        """
        Two handlers of the same priority should still compare.
        """
        handler1 = irc.client.PrioritizedHandler(1, lambda: None)
        handler2 = irc.client.PrioritizedHandler(1, lambda: 'other')
        assert not handler1 < handler2
        assert not handler2 < handler1


@mock.patch('irc.connection.socket')
def test_command_without_arguments(self):
    "A command without arguments should not crash"
    server = irc.client.Reactor().server()
    server.connect('foo', 6667, 'bestnick')
    server._process_line('GLOBALUSERSTATE')



import itertools
import sys

from jaraco.text import clean, drop_comment, lines_from

if sys.version_info >= (3, 12):
    from importlib.resources import files
else:
    from importlib_resources import files


class Command(str):
    def __new__(cls, code, name):
        return super().__new__(cls, name)

    def __init__(self, code, name):
        self.code = code

    def __int__(self):
        return int(self.code)

    @staticmethod
    def lookup(raw) -> 'Command':
        """
        Lookup a command by numeric or by name.

        >>> Command.lookup('002')
        'yourhost'
        >>> Command.lookup('002').code
        '002'
        >>> int(Command.lookup('002'))
        2
        >>> int(Command.lookup('yourhost'))
        2
        >>> Command.lookup('yourhost').code
        '002'

        If a command is supplied that's an unrecognized name or code,
        a Command object is still returned.
        >>> fallback = Command.lookup('Unknown-command')
        >>> fallback
        'unknown-command'
        >>> fallback.code
        'unknown-command'
        >>> int(fallback)
        Traceback (most recent call last):
        ...
        ValueError: invalid literal for int() with base 10: 'unknown-command'
        >>> fallback = Command.lookup('999')
        >>> fallback
        '999'
        >>> int(fallback)
        999
        """
        fallback = Command(raw.lower(), raw.lower())
        return numeric.get(raw, _by_name.get(raw.lower(), fallback))


_codes = itertools.starmap(
    Command,
    map(
        str.split,
        map(drop_comment, clean(lines_from(files().joinpath('codes.txt')))),
    ),
)


numeric = {code.code: code for code in _codes}

codes = {v: k for k, v in numeric.items()}

_by_name = {v: v for v in numeric.values()}

generated = [
    "dcc_connect",
    "dcc_disconnect",
    "dccmsg",
    "disconnect",
    "ctcp",
    "ctcpreply",
    "login_failed",
]

protocol = [
    "error",
    "join",
    "kick",
    "mode",
    "part",
    "ping",
    "privmsg",
    "privnotice",
    "pubmsg",
    "pubnotice",
    "quit",
    "invite",
    "pong",
    "action",
    "topic",
    "nick",
]

all = generated + protocol + list(numeric.values())



from jaraco.text import FoldedCase


class IRCFoldedCase(FoldedCase):
    """
    A version of FoldedCase that honors the IRC specification for lowercased
    strings (RFC 1459).

    >>> IRCFoldedCase('Foo^').lower()
    'foo~'

    >>> IRCFoldedCase('[this]') == IRCFoldedCase('{THIS}')
    True

    >>> IRCFoldedCase('[This]').casefold()
    '{this}'

    >>> IRCFoldedCase().lower()
    ''
    """

    translation = dict(
        zip(
            map(ord, r"[]\^"),
            map(ord, r"{}|~"),
        )
    )

    def lower(self):
        return super().lower().translate(self.translation)

    def casefold(self):
        """
        Ensure cached superclass value doesn't supersede.

        >>> ob = IRCFoldedCase('[This]')
        >>> ob.casefold()
        '{this}'
        >>> ob.casefold()
        '{this}'
        """
        return super().casefold().translate(self.translation)

    def __setattr__(self, key, val):
        if key == 'casefold':
            return
        return super().__setattr__(key, val)


def lower(str):
    return IRCFoldedCase(str).lower()



import socketserver
import threading
import time

import pytest

import irc.bot
import irc.client
import irc.server
from irc.bot import ServerSpec


class TestServerSpec:
    def test_with_host(self):
        server_spec = ServerSpec('irc.example.com')
        assert server_spec.host == 'irc.example.com'
        assert server_spec.port == 6667
        assert server_spec.password is None

    def test_with_host_and_port(self):
        server_spec = ServerSpec('irc.example.org', port=6669)
        assert server_spec.host == 'irc.example.org'
        assert server_spec.port == 6669
        assert server_spec.password is None

    def test_with_host_and_password(self):
        server_spec = ServerSpec('irc.example.net', password='heres johnny!')
        assert server_spec.host == 'irc.example.net'
        assert server_spec.port == 6667
        assert server_spec.password == 'heres johnny!'

    def test_with_host_and_port_and_password(self):
        server_spec = ServerSpec(
            'irc.example.gov', port=6668, password='there-is-only-zuul'
        )
        assert server_spec.host == 'irc.example.gov'
        assert server_spec.port == 6668
        assert server_spec.password == 'there-is-only-zuul'


class TestChannel:
    def test_add_remove_nick(self):
        channel = irc.bot.Channel()
        channel.add_user('tester1')
        channel.remove_user('tester1')
        assert 'tester1' not in channel.users()
        channel.add_user('tester1')
        assert 'tester1' in channel.users()

    def test_change_nick(self):
        channel = irc.bot.Channel()
        channel.add_user('tester1')
        channel.change_nick('tester1', 'was_tester')

    def test_has_user(self):
        channel = irc.bot.Channel()
        channel.add_user('tester1')
        assert channel.has_user('Tester1')

    def test_set_mode_clear_mode(self):
        channel = irc.bot.Channel()
        channel.add_user('tester1')
        channel.set_mode('o', 'tester1')
        assert channel.is_oper('tester1')
        channel.clear_mode('o', 'tester1')
        assert not channel.is_oper('tester1')

    def test_remove_add_clears_mode(self):
        channel = irc.bot.Channel()
        channel.add_user('tester1')
        channel.set_mode('v', 'tester1')
        assert channel.is_voiced('tester1')
        channel.remove_user('tester1')
        channel.add_user('tester1')
        assert not channel.is_voiced('tester1')


class DisconnectHandler(socketserver.BaseRequestHandler):
    """
    Immediately disconnect the client after connecting
    """

    def handle(self):
        self.request.close()


@pytest.fixture
def disconnecting_server():
    """
    An IRC server that disconnects the client immediately.
    """
    # bind to localhost on an ephemeral port
    bind_address = 'localhost', 0
    try:
        srv = irc.server.IRCServer(bind_address, DisconnectHandler)
        threading.Thread(target=srv.serve_forever).start()
        yield srv
    finally:
        srv.shutdown()
        srv.server_close()


class TestBot:
    def test_construct_bot(self):
        bot = irc.bot.SingleServerIRCBot(
            server_list=[('localhost', '9999')],
            realname='irclibbot',
            nickname='irclibbot',
        )
        svr = bot.servers.peek()
        assert svr.host == 'localhost'
        assert svr.port == '9999'
        assert svr.password is None

    def test_namreply_no_channel(self):
        """
        If channel is '*', _on_namreply should not crash.

        Regression test for #22
        """
        event = irc.client.Event(
            type=None, source=None, target=None, arguments=['*', '*', 'nick']
        )
        irc.bot.SingleServerIRCBot._on_namreply(None, None, event)

    def test_reconnects_are_stable(self, disconnecting_server):
        """
        Ensure that disconnects from the server don't lead to
        exponential growth in reconnect attempts.
        """
        recon = irc.bot.ExponentialBackoff(min_interval=0.01)
        bot = irc.bot.SingleServerIRCBot(
            server_list=[disconnecting_server.socket.getsockname()],
            realname='reconnect_test',
            nickname='reconnect_test',
            recon=recon,
        )
        bot._connect()
        for _ in range(4):
            bot.reactor.process_once()
            time.sleep(0.01)
        assert len(bot.reactor.scheduler.queue) <= 1


def test_version():
    assert isinstance(irc.bot.SingleServerIRCBot.get_version(), str)



#! /usr/bin/env python
#
# Example program using irc.client.
#
# This program is free without restrictions; do anything you like with
# it.
#
# Joel Rosdahl <joel@rosdahl.net>


import argparse
import os
import shlex
import struct
import sys

import jaraco.logging

import irc.client


class DCCReceive(irc.client.SimpleIRCClient):
    def __init__(self):
        irc.client.SimpleIRCClient.__init__(self)
        self.received_bytes = 0

    def on_ctcp(self, connection, event):
        payload = event.arguments[1]
        parts = shlex.split(payload)
        command, filename, peer_address, peer_port, size = parts
        if command != "SEND":
            return
        self.filename = os.path.basename(filename)
        if os.path.exists(self.filename):
            print("A file named", self.filename, "already exists. Refusing to save it.")
            self.connection.quit()
            return
        self.file = open(self.filename, "wb")
        peer_address = irc.client.ip_numstr_to_quad(peer_address)
        peer_port = int(peer_port)
        self.dcc = self.dcc_connect(peer_address, peer_port, "raw")

    def on_dccmsg(self, connection, event):
        data = event.arguments[0]
        self.file.write(data)
        self.received_bytes = self.received_bytes + len(data)
        self.dcc.send_bytes(struct.pack("!I", self.received_bytes))

    def on_dcc_disconnect(self, connection, event):
        self.file.close()
        print(f"Received file {self.filename} ({self.received_bytes} bytes).")
        self.connection.quit()

    def on_disconnect(self, connection, event):
        sys.exit(0)


def get_args():
    parser = argparse.ArgumentParser(
        description="Receive a single file to the current directory via DCC "
        "and then exit."
    )
    parser.add_argument('server')
    parser.add_argument('nickname')
    parser.add_argument('-p', '--port', default=6667, type=int)
    jaraco.logging.add_arguments(parser)
    return parser.parse_args()


def main():
    args = get_args()
    jaraco.logging.setup(args)

    c = DCCReceive()
    try:
        c.connect(args.server, args.port, args.nickname)
    except irc.client.ServerConnectionError as x:
        print(x)
        sys.exit(1)
    c.start()


if __name__ == "__main__":
    main()



import abc

from tempora import schedule


class IScheduler(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def execute_every(self, period, func):
        "execute func every period"

    @abc.abstractmethod
    def execute_at(self, when, func):
        "execute func at when"

    @abc.abstractmethod
    def execute_after(self, delay, func):
        "execute func after delay"

    @abc.abstractmethod
    def run_pending(self):
        "invoke the functions that are due"


class DefaultScheduler(schedule.InvokeScheduler, IScheduler):
    def execute_every(self, period, func):
        """
        Executes `func` every `period`.

        :param `func`: function to execute
        :param `period`: `int` in seconds, or `datetime.timedelta`
        """
        self.add(schedule.PeriodicCommand.after(period, func))

    def execute_at(self, when, func):
        self.add(schedule.DelayedCommand.at_time(when, func))

    def execute_after(self, delay, func):
        self.add(schedule.DelayedCommand.after(delay, func))



import re


def get_pages(filename):
    with open(filename) as f:
        data = f.read()
    return data.split('\x0c')


header_pattern = re.compile(r'^RFC \d+\s+.*\s+(\w+ \d{4})$', re.M)
footer_pattern = re.compile(r'^\w+\s+\w+\s+\[Page \d+\]$', re.M)


def remove_header(page):
    page = header_pattern.sub('', page)
    return page.lstrip('\n')


def remove_footer(page):
    page = footer_pattern.sub('', page)
    return page.rstrip() + '\n\n'


def clean_pages():
    return map(remove_header, map(remove_footer, get_pages('rfc2812.txt')))


def save_clean():
    with open('rfc2812-clean.txt', 'w') as f:
        map(f.write, clean_pages())


if __name__ == '__main__':
    save_clean()



import contextlib

try:
    from importlib import metadata  # type: ignore
except ImportError:
    import importlib_metadata as metadata  # type: ignore


def _get_version():
    with contextlib.suppress(Exception):
        return metadata.version('irc')
    return 'unknown'



import asyncio
import contextlib
import warnings
from unittest.mock import MagicMock

from irc import client_aio


def make_mocked_create_connection(mock_transport, mock_protocol):
    async def mock_create_connection(*args, **kwargs):
        return (mock_transport, mock_protocol)

    return mock_create_connection


@contextlib.contextmanager
def suppress_issue_197():
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', 'There is no current event loop')
        yield


def test_privmsg_sends_msg():
    # create dummy transport, protocol
    mock_transport = MagicMock()
    mock_protocol = MagicMock()

    # connect to dummy server
    with suppress_issue_197():
        loop = asyncio.new_event_loop()
    loop.create_connection = make_mocked_create_connection(
        mock_transport, mock_protocol
    )
    server = client_aio.AioReactor(loop=loop).server()
    loop.run_until_complete(server.connect('foo', 6667, 'my_irc_nick'))
    server.privmsg('#best-channel', 'You are great')

    mock_transport.write.assert_called_with(b'PRIVMSG #best-channel :You are great\r\n')

    loop.close()



"""
Simple IRC bot library.

This module contains a single-server IRC bot class that can be used to
write simpler bots.
"""

import abc
import collections
import itertools
import random
import sys

import more_itertools

import irc.client
import irc.modes

from .dict import IRCDict


class ServerSpec:
    """
    An IRC server specification.

    >>> spec = ServerSpec('localhost')
    >>> spec.host
    'localhost'
    >>> spec.port
    6667
    >>> spec.password

    >>> spec = ServerSpec('127.0.0.1', 6697, 'fooP455')
    >>> spec.password
    'fooP455'
    """

    def __init__(self, host, port=6667, password=None):
        self.host = host
        self.port = port
        self.password = password

    def __repr__(self):
        return "<irc.bot.ServerSpec for server {}:{} {}>".format(
            self.host,
            self.port,
            "with password" if self.password else "without password",
        )

    @classmethod
    def ensure(cls, input):
        spec = cls(*input) if isinstance(input, (list, tuple)) else input
        assert isinstance(spec, cls)
        return spec


class ReconnectStrategy(metaclass=abc.ABCMeta):
    """
    An abstract base class describing the interface used by
    SingleServerIRCBot for handling reconnect following
    disconnect events.
    """

    @abc.abstractmethod
    def run(self, bot):
        """
        Invoked by the bot on disconnect. Here
        a strategy can determine how to react to a
        disconnect.
        """


class ExponentialBackoff(ReconnectStrategy):
    """
    A ReconnectStrategy implementing exponential backoff
    with jitter.
    """

    min_interval = 60
    max_interval = 300

    def __init__(self, **attrs):
        vars(self).update(attrs)
        assert 0 <= self.min_interval <= self.max_interval
        self._check_scheduled = False
        self.attempt_count = itertools.count(1)

    def run(self, bot):
        self.bot = bot

        if self._check_scheduled:
            return

        # calculate interval in seconds based on connection attempts
        intvl = 2 ** next(self.attempt_count) - 1

        # limit the max interval
        intvl = min(intvl, self.max_interval)

        # add jitter and truncate to integer seconds
        intvl = int(intvl * random.random())

        # limit the min interval
        intvl = max(intvl, self.min_interval)

        self.bot.reactor.scheduler.execute_after(intvl, self.check)
        self._check_scheduled = True

    def check(self):
        self._check_scheduled = False
        if not self.bot.connection.is_connected():
            self.run(self.bot)
            self.bot.jump_server()


missing = object()


class SingleServerIRCBot(irc.client.SimpleIRCClient):
    r"""A single-server IRC bot class.

    The bot tries to reconnect if it is disconnected.

    The bot keeps track of the channels it has joined, the other
    clients that are present in the channels and which of those that
    have operator or voice modes.  The "database" is kept in the
    self.channels attribute, which is an IRCDict of Channels.

    Arguments:

        server_list -- A list of ServerSpec objects or tuples of
            parameters suitable for constructing ServerSpec
            objects. Defines the list of servers the bot will
            use (in order).

        nickname -- The bot's nickname.

        realname -- The bot's realname.

        recon -- A ReconnectStrategy for reconnecting on
            disconnect or failed connection.

        dcc_connections -- A list of initiated/accepted DCC
            connections.

        \*\*connect_params -- parameters to pass through to the connect
            method.
    """

    def __init__(
        self,
        server_list,
        nickname,
        realname,
        _=None,
        recon=ExponentialBackoff(),
        **connect_params,
    ):
        super().__init__()
        self.__connect_params = connect_params
        self.channels = IRCDict()
        specs = map(ServerSpec.ensure, server_list)
        self.servers = more_itertools.peekable(itertools.cycle(specs))
        self.recon = recon

        self._nickname = nickname
        self._realname = realname
        for i in [
            "disconnect",
            "join",
            "kick",
            "mode",
            "namreply",
            "nick",
            "part",
            "quit",
        ]:
            self.connection.add_global_handler(i, getattr(self, "_on_" + i), -20)

    def _connect(self):
        """
        Establish a connection to the server at the front of the server_list.
        """
        server = self.servers.peek()
        try:
            self.connect(
                server.host,
                server.port,
                self._nickname,
                server.password,
                ircname=self._realname,
                **self.__connect_params,
            )
        except irc.client.ServerConnectionError:
            self.connection._handle_event(
                irc.client.Event("disconnect", self.connection.server, "", [""])
            )

    def _on_disconnect(self, connection, event):
        self.channels = IRCDict()
        self.recon.run(self)

    def _on_join(self, connection, event):
        ch = event.target
        nick = event.source.nick
        if nick == connection.get_nickname():
            self.channels[ch] = Channel()
        self.channels[ch].add_user(nick)

    def _on_kick(self, connection, event):
        nick = event.arguments[0]
        channel = event.target

        if nick == connection.get_nickname():
            del self.channels[channel]
        else:
            self.channels[channel].remove_user(nick)

    def _on_mode(self, connection, event):
        t = event.target
        if not irc.client.is_channel(t):
            # mode on self; disregard
            return
        ch = self.channels[t]

        modes = irc.modes.parse_channel_modes(" ".join(event.arguments))
        for sign, mode, argument in modes:
            f = {"+": ch.set_mode, "-": ch.clear_mode}[sign]
            f(mode, argument)

    def _on_namreply(self, connection, event):
        """
        event.arguments[0] == "@" for secret channels,
                          "*" for private channels,
                          "=" for others (public channels)
        event.arguments[1] == channel
        event.arguments[2] == nick list
        """

        ch_type, channel, nick_list = event.arguments

        if channel == '*':
            # User is not in any visible channel
            # http://tools.ietf.org/html/rfc2812#section-3.2.5
            return

        for nick in nick_list.split():
            nick_modes = []

            if nick[0] in self.connection.features.prefix:
                nick_modes.append(self.connection.features.prefix[nick[0]])
                nick = nick[1:]

            for mode in nick_modes:
                self.channels[channel].set_mode(mode, nick)

            self.channels[channel].add_user(nick)

    def _on_nick(self, connection, event):
        before = event.source.nick
        after = event.target
        for ch in self.channels.values():
            if ch.has_user(before):
                ch.change_nick(before, after)

    def _on_part(self, connection, event):
        nick = event.source.nick
        channel = event.target

        if nick == connection.get_nickname():
            del self.channels[channel]
        else:
            self.channels[channel].remove_user(nick)

    def _on_quit(self, connection, event):
        nick = event.source.nick
        for ch in self.channels.values():
            if ch.has_user(nick):
                ch.remove_user(nick)

    def die(self, msg="Bye, cruel world!"):
        """Let the bot die.

        Arguments:

            msg -- Quit message.
        """

        self.connection.disconnect(msg)
        sys.exit(0)

    def disconnect(self, msg="I'll be back!"):
        """Disconnect the bot.

        The bot will try to reconnect after a while.

        Arguments:

            msg -- Quit message.
        """
        self.connection.disconnect(msg)

    @staticmethod
    def get_version():
        """Returns the bot version.

        Used when answering a CTCP VERSION request.
        """
        return f"Python irc.bot ({irc._get_version()})"

    def jump_server(self, msg="Changing servers"):
        """Connect to a new server, possibly disconnecting from the current.

        The bot will skip to next server in the server_list each time
        jump_server is called.
        """
        if self.connection.is_connected():
            self.connection.disconnect(msg)

        next(self.servers)
        self._connect()

    def on_ctcp(self, connection, event):
        """Default handler for ctcp events.

        Replies to VERSION and PING requests and relays DCC requests
        to the on_dccchat method.
        """
        nick = event.source.nick
        if event.arguments[0] == "VERSION":
            connection.ctcp_reply(nick, "VERSION " + self.get_version())
        elif event.arguments[0] == "PING":
            if len(event.arguments) > 1:
                connection.ctcp_reply(nick, "PING " + event.arguments[1])
        elif (
            event.arguments[0] == "DCC"
            and event.arguments[1].split(" ", 1)[0] == "CHAT"
        ):
            self.on_dccchat(connection, event)

    def on_dccchat(self, connection, event):
        pass

    def start(self):
        """Start the bot."""
        self._connect()
        super().start()


class Channel:
    """
    A class for keeping information about an IRC channel.
    """

    user_modes = 'ovqha'
    """
    Modes which are applicable to individual users, and which
    should be tracked in the mode_users dictionary.
    """

    def __init__(self):
        self._users = IRCDict()
        self.mode_users = collections.defaultdict(IRCDict)
        self.modes = {}

    def users(self):
        """Returns an unsorted list of the channel's users."""
        return self._users.keys()

    def opers(self):
        """Returns an unsorted list of the channel's operators."""
        return self.mode_users['o'].keys()

    def voiced(self):
        """Returns an unsorted list of the persons that have voice
        mode set in the channel."""
        return self.mode_users['v'].keys()

    def owners(self):
        """Returns an unsorted list of the channel's owners."""
        return self.mode_users['q'].keys()

    def halfops(self):
        """Returns an unsorted list of the channel's half-operators."""
        return self.mode_users['h'].keys()

    def admins(self):
        """Returns an unsorted list of the channel's admins."""
        return self.mode_users['a'].keys()

    def has_user(self, nick):
        """Check whether the channel has a user."""
        return nick in self._users

    def is_oper(self, nick):
        """Check whether a user has operator status in the channel."""
        return nick in self.mode_users['o']

    def is_voiced(self, nick):
        """Check whether a user has voice mode set in the channel."""
        return nick in self.mode_users['v']

    def is_owner(self, nick):
        """Check whether a user has owner status in the channel."""
        return nick in self.mode_users['q']

    def is_halfop(self, nick):
        """Check whether a user has half-operator status in the channel."""
        return nick in self.mode_users['h']

    def is_admin(self, nick):
        """Check whether a user has admin status in the channel."""
        return nick in self.mode_users['a']

    def add_user(self, nick):
        self._users[nick] = 1

    @property
    def user_dicts(self):
        yield self._users
        yield from self.mode_users.values()

    def remove_user(self, nick):
        for d in self.user_dicts:
            d.pop(nick, None)

    def change_nick(self, before, after):
        self._users[after] = self._users.pop(before)
        for mode_lookup in self.mode_users.values():
            if before in mode_lookup:
                mode_lookup[after] = mode_lookup.pop(before)

    def set_userdetails(self, nick, details):
        if nick in self._users:
            self._users[nick] = details

    def set_mode(self, mode, value=None):
        """Set mode on the channel.

        Arguments:

            mode -- The mode (a single-character string).

            value -- Value
        """
        if mode in self.user_modes:
            self.mode_users[mode][value] = 1
        else:
            self.modes[mode] = value

    def clear_mode(self, mode, value=None):
        """Clear mode on the channel.

        Arguments:

            mode -- The mode (a single-character string).

            value -- Value
        """
        try:
            if mode in self.user_modes:
                del self.mode_users[mode][value]
            else:
                del self.modes[mode]
        except KeyError:
            pass

    def has_mode(self, mode):
        return mode in self.modes

    def is_moderated(self):
        return self.has_mode("m")

    def is_secret(self):
        return self.has_mode("s")

    def is_protected(self):
        return self.has_mode("p")

    def has_topic_lock(self):
        return self.has_mode("t")

    def is_invite_only(self):
        return self.has_mode("i")

    def has_allow_external_messages(self):
        return self.has_mode("n")

    def has_limit(self):
        return self.has_mode("l")

    def limit(self):
        if self.has_limit():
            return self.modes["l"]
        else:
            return None

    def has_key(self):
        return self.has_mode("k")



"""
Internet Relay Chat (IRC) protocol client library.

This library is intended to encapsulate the IRC protocol in Python.
It provides an event-driven IRC client framework.  It has
a fairly thorough support for the basic IRC protocol, CTCP, and DCC chat.

To best understand how to make an IRC client, the reader more
or less must understand the IRC specifications.  They are available
here: [IRC specifications].

The main features of the IRC client framework are:

  * Abstraction of the IRC protocol.
  * Handles multiple simultaneous IRC server connections.
  * Handles server PONGing transparently.
  * Messages to the IRC server are done by calling methods on an IRC
    connection object.
  * Messages from an IRC server triggers events, which can be caught
    by event handlers.
  * Reading from and writing to IRC server sockets are normally done
    by an internal select() loop, but the select()ing may be done by
    an external main loop.
  * Functions can be registered to execute at specified times by the
    event-loop.
  * Decodes CTCP tagging correctly (hopefully); I haven't seen any
    other IRC client implementation that handles the CTCP
    specification subtleties.
  * A kind of simple, single-server, object-oriented IRC client class
    that dispatches events to instance methods is included.

Current limitations:

  * Data is not written asynchronously to the server, i.e. the write()
    may block if the TCP buffers are stuffed.
  * DCC file transfers are not supported.
  * RFCs 2810, 2811, 2812, and 2813 have not been considered.

Notes:
  * connection.quit() only sends QUIT to the server.
  * ERROR from the server triggers the error event and the disconnect event.
  * dropping of the connection triggers the disconnect event.


.. [IRC specifications] http://www.irchelp.org/irchelp/rfc/
"""

from __future__ import annotations

import abc
import base64
import bisect
import collections
import contextlib
import functools
import itertools
import logging
import re
import select
import socket
import struct
import threading
import time
import warnings
from typing import Callable

import jaraco.functools
from jaraco.functools import Throttler
from jaraco.stream import buffer
from more_itertools import always_iterable, consume, repeatfunc

from . import connection, ctcp, events, features, message, schedule

log = logging.getLogger(__name__)


class IRCError(Exception):
    "An IRC exception"


class InvalidCharacters(ValueError):
    "Invalid characters were encountered in the message"


class MessageTooLong(ValueError):
    "Message is too long"


class Connection(metaclass=abc.ABCMeta):
    """
    Base class for IRC connections.
    """

    transmit_encoding = 'utf-8'
    "encoding used for transmission"

    @abc.abstractproperty
    def socket(self):
        "The socket for this connection"

    def __init__(self, reactor):
        self.reactor = reactor

    def encode(self, msg):
        """Encode a message for transmission."""
        return msg.encode(self.transmit_encoding)


class ServerConnectionError(IRCError):
    pass


class ServerNotConnectedError(ServerConnectionError):
    pass


class ServerConnection(Connection):
    """
    An IRC server connection.

    ServerConnection objects are instantiated by calling the server
    method on a Reactor object.
    """

    buffer_class = buffer.DecodingLineBuffer
    socket = None
    connected = False

    def __init__(self, reactor):
        super().__init__(reactor)
        self.features = features.FeatureSet()

    # save the method args to allow for easier reconnection.
    @jaraco.functools.save_method_args
    def connect(
        self,
        server: str,
        port: int,
        nickname: str,
        password: str | None = None,
        username: str | None = None,
        ircname: str | None = None,
        connect_factory=connection.Factory(),
        sasl_login: str | None = None,
    ):
        """Connect/reconnect to a server.

        Arguments:

        * server - Server name
        * port - Port number
        * nickname - The nickname
        * password - Password (if any)
        * username - The username
        * ircname - The IRC name ("realname")
        * server_address - The remote host/port of the server
        * connect_factory - A callable that takes the server address and
          returns a connection (with a socket interface)
        * sasl_login - A string used to toggle sasl plain login.
          Password needs to be set as well, and will be used for SASL,
          not PASS login.

        This function can be called to reconnect a closed connection.

        Returns the ServerConnection object.
        """
        log.debug(
            "connect(server=%r, port=%r, nickname=%r, ...)", server, port, nickname
        )

        if self.connected:
            self.disconnect("Changing servers")

        self.buffer = self.buffer_class()
        self.handlers: dict[str, Callable] = {}
        self.real_server_name = ""
        self.real_nickname = nickname
        self.server = server
        self.port = port
        self.server_address = (server, port)
        self.nickname = nickname
        self.username = username or nickname
        self.ircname = ircname or nickname
        self.password = password
        self.connect_factory = connect_factory
        self.sasl_login = sasl_login
        try:
            self.socket = self.connect_factory(self.server_address)
        except OSError as ex:
            raise ServerConnectionError(f"Couldn't connect to socket: {ex}") from ex
        self.connected = True
        self.reactor._on_connect(self.socket)

        # Log on...
        if self.sasl_login and self.password:
            self._sasl_step = None
            for i in ["cap", "authenticate", "saslsuccess", "saslfail"]:
                self.add_global_handler(i, self._sasl_state_machine, -42)
            self.cap("LS")
            self.nick(self.nickname)
            self.user(self.username, self.ircname)
            self._sasl_step = self._sasl_cap_ls
            return self
        if self.password:
            self.pass_(self.password)
        self.nick(self.nickname)
        self.user(self.username, self.ircname)
        return self

    def _sasl_state_machine(self, connection, event):
        if self._sasl_step:
            self._sasl_step(event)

    def _sasl_cap_ls(self, event):
        if (
            event.type == "cap"
            and len(event.arguments) > 1
            and event.arguments[0] == "LS"
        ):
            if 'sasl' in event.arguments[1].split():
                self.cap("REQ", "sasl")
                self._sasl_step = self._sasl_cap_req
            else:
                event = Event(
                    "login_failed", event.target, ["server does not support sasl"]
                )
                self._handle_event(event)
                self._sasl_end()

    def _sasl_cap_req(self, event):
        if event.type == "cap" and len(event.arguments) > 1:
            if event.arguments[0] == "ACK" and 'sasl' in event.arguments:
                self.send_items('AUTHENTICATE', 'PLAIN')
                self._sasl_step = self._sasl_auth_plain
            elif event.arguments[0] == "NAK":
                event = Event(
                    "login_failed", event.target, ["server refused sasl protocol"]
                )
                self._handle_event(event)
                self._sasl_end()

    def _sasl_auth_plain(self, event):
        if event.type == "authenticate" and event.target == "+":
            auth_string = base64.b64encode(
                self.encode(f"\x00{self.sasl_login}\x00{self.password}")
            ).decode()
            self.send_items('AUTHENTICATE', auth_string)
            self._sasl_step = self._sasl_auth_sent

    def _sasl_auth_sent(self, event):
        if event.type == "saslsuccess":
            self._sasl_end()
        elif event.type == "saslfail":
            event = Event("login_failed", event.target, event.arguments)
            self._handle_event(event)
            self._sasl_end()

    def _sasl_end(self):
        self._sasl_step = None
        self.cap("END")
        # SASL done, de-register handlers
        for i in ["cap", "authenticate", "saslsuccess", "saslfail"]:
            self.remove_global_handler(i, self._sasl_state_machine)

    def reconnect(self):
        """
        Reconnect with the last arguments passed to self.connect()
        """
        self.connect(*self._saved_connect.args, **self._saved_connect.kwargs)

    def close(self):
        """Close the connection.

        This method closes the connection permanently; after it has
        been called, the object is unusable.
        """
        # Without this thread lock, there is a window during which
        # select() can find a closed socket, leading to an EBADF error.
        with self.reactor.mutex:
            self.disconnect("Closing object")
            self.reactor._remove_connection(self)

    def get_server_name(self):
        """Get the (real) server name.

        This method returns the (real) server name, or, more
        specifically, what the server calls itself.
        """
        return self.real_server_name or ""

    def get_nickname(self):
        """Get the (real) nick name.

        This method returns the (real) nickname.  The library keeps
        track of nick changes, so it might not be the nick name that
        was passed to the connect() method.
        """
        return self.real_nickname

    @contextlib.contextmanager
    def as_nick(self, name):
        """
        Set the nick for the duration of the context.
        """
        orig = self.get_nickname()
        self.nick(name)
        try:
            yield orig
        finally:
            self.nick(orig)

    def process_data(self):
        "read and process input from self.socket"

        try:
            reader = getattr(self.socket, 'read', self.socket.recv)
            new_data = reader(2**14)
        except OSError:
            # The server hung up.
            self.disconnect("Connection reset by peer")
            return
        if not new_data:
            # Read nothing: connection must be down.
            self.disconnect("Connection reset by peer")
            return

        self.buffer.feed(new_data)

        # process each non-empty line after logging all lines
        for line in self.buffer:
            log.debug("FROM SERVER: %s", line)
            if not line:
                continue
            self._process_line(line)

    def _process_line(self, line):
        event = Event("all_raw_messages", self.get_server_name(), None, [line])
        self._handle_event(event)

        grp = _rfc_1459_command_regexp.match(line).group

        source = NickMask.from_group(grp("prefix"))
        command = events.Command.lookup(grp("command"))
        arguments = message.Arguments.from_group(grp('argument'))
        tags = message.Tag.from_group(grp('tags'))

        if source and not self.real_server_name:
            self.real_server_name = source

        if command == "nick":
            if source.nick == self.real_nickname:
                self.real_nickname = arguments[0]
        elif command == "welcome":
            # Record the nickname in case the client changed nick
            # in a nicknameinuse callback.
            self.real_nickname = arguments[0]
        elif command == "featurelist":
            self.features.load(arguments)

        handler = (
            self._handle_message
            if command in ["privmsg", "notice"]
            else self._handle_other
        )
        handler(arguments, command, source, tags)

    def _handle_message(self, arguments, command, source, tags):
        target, msg = arguments[:2]
        messages = ctcp.dequote(msg)
        if command == "privmsg":
            if is_channel(target):
                command = "pubmsg"
        else:
            if is_channel(target):
                command = "pubnotice"
            else:
                command = "privnotice"
        for m in messages:
            if isinstance(m, tuple):
                if command in ["privmsg", "pubmsg"]:
                    command = "ctcp"
                else:
                    command = "ctcpreply"

                m = list(m)
                log.debug(
                    "command: %s, source: %s, target: %s, " "arguments: %s, tags: %s",
                    command,
                    source,
                    target,
                    m,
                    tags,
                )
                event = Event(command, source, target, m, tags)
                self._handle_event(event)
                if command == "ctcp" and m[0] == "ACTION":
                    event = Event("action", source, target, m[1:], tags)
                    self._handle_event(event)
            else:
                log.debug(
                    "command: %s, source: %s, target: %s, " "arguments: %s, tags: %s",
                    command,
                    source,
                    target,
                    [m],
                    tags,
                )
                event = Event(command, source, target, [m], tags)
                self._handle_event(event)

    def _handle_other(self, arguments, command, source, tags):
        target = None
        if command == "quit":
            arguments = [arguments[0]]
        elif command == "ping":
            target = arguments[0]
        else:
            target = arguments[0] if arguments else None
            arguments = arguments[1:]
        if command == "mode":
            if not is_channel(target):
                command = "umode"
        log.debug(
            "command: %s, source: %s, target: %s, " "arguments: %s, tags: %s",
            command,
            source,
            target,
            arguments,
            tags,
        )
        event = Event(command, source, target, arguments, tags)
        self._handle_event(event)

    def _handle_event(self, event):
        """[Internal]"""
        self.reactor._handle_event(self, event)
        if event.type in self.handlers:
            for fn in self.handlers[event.type]:
                fn(self, event)

    def is_connected(self):
        """Return connection status.

        Returns true if connected, otherwise false.
        """
        return self.connected

    def add_global_handler(self, *args):
        """Add global handler.

        See documentation for IRC.add_global_handler.
        """
        self.reactor.add_global_handler(*args)

    def remove_global_handler(self, *args):
        """Remove global handler.

        See documentation for IRC.remove_global_handler.
        """
        self.reactor.remove_global_handler(*args)

    def action(self, target, action):
        """Send a CTCP ACTION command."""
        self.ctcp("ACTION", target, action)

    def admin(self, server=""):
        """Send an ADMIN command."""
        self.send_items('ADMIN', server)

    def cap(self, subcommand, *args):
        """
        Send a CAP command according to `the spec
        <http://ircv3.atheme.org/specification/capability-negotiation-3.1>`_.

        Arguments:

            subcommand -- LS, LIST, REQ, ACK, CLEAR, END
            args -- capabilities, if required for given subcommand

        Example:

            .cap('LS')
            .cap('REQ', 'multi-prefix', 'sasl')
            .cap('END')
        """
        cap_subcommands = set('LS LIST REQ ACK NAK CLEAR END'.split())
        client_subcommands = set(cap_subcommands) - {'NAK'}
        assert subcommand in client_subcommands, "invalid subcommand"

        def _multi_parameter(args):
            """
            According to the spec::

                If more than one capability is named, the RFC1459 designated
                sentinel (:) for a multi-parameter argument must be present.

            It's not obvious where the sentinel should be present or if it
            must be omitted for a single parameter, so follow convention and
            only include the sentinel prefixed to the first parameter if more
            than one parameter is present.
            """
            if len(args) > 1:
                return (':' + args[0],) + args[1:]
            return args

        self.send_items('CAP', subcommand, *_multi_parameter(args))

    def ctcp(self, ctcptype, target, parameter=""):
        """Send a CTCP command."""
        ctcptype = ctcptype.upper()
        self.privmsg(
            target,
            f"\001{ctcptype} {parameter}\001" if parameter else "\001{ctcptype}\001",
        )

    def ctcp_reply(self, target, parameter):
        """Send a CTCP REPLY command."""
        self.notice(target, f"\001{parameter}\001")

    def disconnect(self, message=""):
        """Hang up the connection.

        Arguments:

            message -- Quit message.
        """
        try:
            del self.connected
        except AttributeError:
            return

        self.quit(message)

        try:
            self.socket.shutdown(socket.SHUT_WR)
        except OSError:
            pass

        self.socket.close()

        del self.socket
        self._handle_event(Event("disconnect", self.server, "", [message]))

    def globops(self, text):
        """Send a GLOBOPS command."""
        self.send_items('GLOBOPS', ':' + text)

    def info(self, server=""):
        """Send an INFO command."""
        self.send_items('INFO', server)

    def invite(self, nick, channel):
        """Send an INVITE command."""
        self.send_items('INVITE', nick, channel)

    def ison(self, nicks):
        """Send an ISON command.

        Arguments:

            nicks -- List of nicks.
        """
        self.send_items('ISON', *tuple(nicks))

    def join(self, channel, key=""):
        """Send a JOIN command."""
        self.send_items('JOIN', channel, key)

    def kick(self, channel, nick, comment=""):
        """Send a KICK command."""
        self.send_items('KICK', channel, nick, comment and ':' + comment)

    def links(self, remote_server="", server_mask=""):
        """Send a LINKS command."""
        self.send_items('LINKS', remote_server, server_mask)

    def list(self, channels=None, server=""):
        """Send a LIST command."""
        self.send_items('LIST', ','.join(always_iterable(channels)), server)

    def lusers(self, server=""):
        """Send a LUSERS command."""
        self.send_items('LUSERS', server)

    def mode(self, target, command):
        """Send a MODE command."""
        self.send_items('MODE', target, command)

    def motd(self, server=""):
        """Send an MOTD command."""
        self.send_items('MOTD', server)

    def names(self, channels=None):
        """Send a NAMES command."""
        self.send_items('NAMES', ','.join(always_iterable(channels)))

    def nick(self, newnick):
        """Send a NICK command."""
        self.send_items('NICK', newnick)

    def notice(self, target, text):
        """Send a NOTICE command."""
        # Should limit len(text) here!
        self.send_items('NOTICE', target, ':' + text)

    def oper(self, nick, password):
        """Send an OPER command."""
        self.send_items('OPER', nick, password)

    def part(self, channels, message=""):
        """Send a PART command."""
        self.send_items('PART', ','.join(always_iterable(channels)), message)

    def pass_(self, password):
        """Send a PASS command."""
        self.send_items('PASS', password)

    def ping(self, target, target2=""):
        """Send a PING command."""
        self.send_items('PING', target, target2)

    def pong(self, target, target2=""):
        """Send a PONG command."""
        self.send_items('PONG', target, target2)

    def privmsg(self, target, text):
        """Send a PRIVMSG command."""
        self.send_items('PRIVMSG', target, ':' + text)

    def privmsg_many(self, targets, text):
        """Send a PRIVMSG command to multiple targets."""
        target = ','.join(targets)
        return self.privmsg(target, text)

    def quit(self, message=""):
        """Send a QUIT command."""
        # Note that many IRC servers don't use your QUIT message
        # unless you've been connected for at least 5 minutes!
        self.send_items('QUIT', message and ':' + message)

    def _prep_message(self, string):
        # The string should not contain any carriage return other than the
        # one added here.
        if '\n' in string:
            msg = "Carriage returns not allowed in privmsg(text)"
            raise InvalidCharacters(msg)
        bytes = self.encode(string) + b'\r\n'
        # According to the RFC http://tools.ietf.org/html/rfc2812#page-6,
        # clients should not transmit more than 512 bytes.
        if len(bytes) > 512:
            msg = "Messages limited to 512 bytes including CR/LF"
            raise MessageTooLong(msg)
        return bytes

    def send_items(self, *items):
        """
        Send all non-empty items, separated by spaces.
        """
        self.send_raw(' '.join(filter(None, items)))

    def send_raw(self, string):
        """Send raw string to the server.

        The string will be padded with appropriate CR LF.
        """
        if self.socket is None:
            raise ServerNotConnectedError("Not connected.")
        sender = getattr(self.socket, 'write', self.socket.send)
        try:
            sender(self._prep_message(string))
            log.debug("TO SERVER: %s", string)
        except OSError:
            # Ouch!
            self.disconnect("Connection reset by peer.")

    def squit(self, server, comment=""):
        """Send an SQUIT command."""
        self.send_items('SQUIT', server, comment and ':' + comment)

    def stats(self, statstype, server=""):
        """Send a STATS command."""
        self.send_items('STATS', statstype, server)

    def time(self, server=""):
        """Send a TIME command."""
        self.send_items('TIME', server)

    def topic(self, channel, new_topic=None):
        """Send a TOPIC command."""
        self.send_items('TOPIC', channel, new_topic and ':' + new_topic)

    def trace(self, target=""):
        """Send a TRACE command."""
        self.send_items('TRACE', target)

    def user(self, username, realname):
        """Send a USER command."""
        cmd = f'USER {username} 0 * :{realname}'
        self.send_raw(cmd)

    def userhost(self, nicks):
        """Send a USERHOST command."""
        self.send_items('USERHOST', ",".join(nicks))

    def users(self, server=""):
        """Send a USERS command."""
        self.send_items('USERS', server)

    def version(self, server=""):
        """Send a VERSION command."""
        self.send_items('VERSION', server)

    def wallops(self, text):
        """Send a WALLOPS command."""
        self.send_items('WALLOPS', ':' + text)

    def who(self, target="", op=""):
        """Send a WHO command."""
        self.send_items('WHO', target, op and 'o')

    def whois(self, targets):
        """Send a WHOIS command."""
        self.send_items('WHOIS', ",".join(always_iterable(targets)))

    def whowas(self, nick, max="", server=""):
        """Send a WHOWAS command."""
        self.send_items('WHOWAS', nick, max, server)

    def set_rate_limit(self, frequency):
        """
        Set a `frequency` limit (messages per second) for this connection.
        Any attempts to send faster than this rate will block.
        """
        self.send_raw = Throttler(self.send_raw, frequency)

    def set_keepalive(self, interval):
        """
        Set a keepalive to occur every `interval` on this `ServerConnection`.

        :param interval: `int` in seconds, or `datetime.timedelta`
        """
        pinger = functools.partial(self.ping, 'keep-alive')
        self.reactor.scheduler.execute_every(period=interval, func=pinger)


class PrioritizedHandler(collections.namedtuple('Base', ('priority', 'callback'))):
    def __lt__(self, other):
        "when sorting prioritized handlers, only use the priority"
        return self.priority < other.priority


class Reactor:
    """
    Processes events from one or more IRC server connections.

    This class implements a reactor in the style of the `reactor pattern
    <http://en.wikipedia.org/wiki/Reactor_pattern>`_.

    When a Reactor object has been instantiated, it can be used to create
    Connection objects that represent the IRC connections.  The
    responsibility of the reactor object is to provide an event-driven
    framework for the connections and to keep the connections alive.
    It runs a select loop to poll each connection's TCP socket and
    hands over the sockets with incoming data for processing by the
    corresponding connection.

    The methods of most interest for an IRC client writer are server,
    add_global_handler, remove_global_handler,
    process_once, and process_forever.

    This is functionally an event-loop which can either use it's own
    internal polling loop, or tie into an external event-loop, by
    having the external event-system periodically call `process_once`
    on the instantiated reactor class. This will allow the reactor
    to process any queued data and/or events.

    Calling `process_forever` will hand off execution to the reactor's
    internal event-loop, which will not return for the life of the
    reactor.

    Here is an example:

        client = irc.client.Reactor()
        server = client.server()
        server.connect("irc.some.where", 6667, "my_nickname")
        server.privmsg("a_nickname", "Hi there!")
        client.process_forever()

    This will connect to the IRC server irc.some.where on port 6667
    using the nickname my_nickname and send the message "Hi there!"
    to the nickname a_nickname.

    The methods of this class are thread-safe; accesses to and modifications
    of its internal lists of connections, handlers, and delayed commands
    are guarded by a mutex.
    """

    scheduler_class = schedule.DefaultScheduler
    connection_class = ServerConnection

    def __do_nothing(*args, **kwargs):
        pass

    def __init__(self, on_connect=__do_nothing, on_disconnect=__do_nothing):
        """Constructor for Reactor objects.

        on_connect: optional callback invoked when a new connection
        is made.

        on_disconnect: optional callback invoked when a socket is
        disconnected.

        The arguments mainly exist to be able to use an external
        main loop (for example Tkinter's or PyGTK's main app loop)
        instead of calling the process_forever method.

        An alternative is to just call ServerConnection.process_once()
        once in a while.
        """

        self._on_connect = on_connect
        self._on_disconnect = on_disconnect

        scheduler = self.scheduler_class()
        assert isinstance(scheduler, schedule.IScheduler)
        self.scheduler = scheduler

        self.connections = []
        self.handlers = {}
        # Modifications to these shared lists and dict need to be thread-safe
        self.mutex = threading.RLock()

        self.add_global_handler("ping", _ping_ponger, -42)

    def server(self):
        """Creates and returns a ServerConnection object."""

        conn = self.connection_class(self)
        with self.mutex:
            self.connections.append(conn)
        return conn

    def process_data(self, sockets):
        """Called when there is more data to read on connection sockets.

        Arguments:

            sockets -- A list of socket objects.

        See documentation for Reactor.__init__.
        """
        with self.mutex:
            log.log(logging.DEBUG - 2, "process_data()")
            for sock, conn in itertools.product(sockets, self.connections):
                if sock == conn.socket:
                    conn.process_data()

    def process_timeout(self):
        """Called when a timeout notification is due.

        See documentation for Reactor.__init__.
        """
        with self.mutex:
            self.scheduler.run_pending()

    @property
    def sockets(self):
        with self.mutex:
            return [
                conn.socket
                for conn in self.connections
                if conn is not None and conn.socket is not None
            ]

    def process_once(self, timeout=0):
        """Process data from connections once.

        Arguments:

            timeout -- How long the select() call should wait if no
                       data is available.

        This method should be called periodically to check and process
        incoming data, if there are any.  If that seems boring, look
        at the process_forever method.
        """
        log.log(logging.DEBUG - 2, "process_once()")
        sockets = self.sockets
        if sockets:
            in_, out, err = select.select(sockets, [], [], timeout)
            self.process_data(in_)
        else:
            time.sleep(timeout)
        self.process_timeout()

    def process_forever(self, timeout=0.2):
        """Run an infinite loop, processing data from connections.

        This method repeatedly calls process_once.

        Arguments:

            timeout -- Parameter to pass to process_once.
        """
        # This loop should specifically *not* be mutex-locked.
        # Otherwise no other thread would ever be able to change
        # the shared state of a Reactor object running this function.
        log.debug("process_forever(timeout=%s)", timeout)
        one = functools.partial(self.process_once, timeout=timeout)
        consume(repeatfunc(one))

    def disconnect_all(self, message=""):
        """Disconnects all connections."""
        with self.mutex:
            for conn in self.connections:
                conn.disconnect(message)

    def add_global_handler(self, event, handler, priority=0):
        """Adds a global handler function for a specific event type.

        Arguments:

            event -- Event type (a string).  Check the values of
                     numeric_events for possible event types.

            handler -- Callback function taking 'connection' and 'event'
                       parameters.

            priority -- A number (the lower number, the higher priority).

        The handler function is called whenever the specified event is
        triggered in any of the connections.  See documentation for
        the Event class.

        The handler functions are called in priority order (lowest
        number is highest priority).  If a handler function returns
        "NO MORE", no more handlers will be called.
        """
        handler = PrioritizedHandler(priority, handler)
        with self.mutex:
            event_handlers = self.handlers.setdefault(event, [])
            bisect.insort(event_handlers, handler)

    def remove_global_handler(self, event, handler):
        """Removes a global handler function.

        Arguments:

            event -- Event type (a string).
            handler -- Callback function.

        Returns 1 on success, otherwise 0.
        """
        with self.mutex:
            if event not in self.handlers:
                return 0
            for h in self.handlers[event]:
                if handler == h.callback:
                    self.handlers[event].remove(h)
        return 1

    def dcc(self, dcctype="chat"):
        """Creates and returns a DCCConnection object.

        Arguments:

            dcctype -- "chat" for DCC CHAT connections or "raw" for
                       DCC SEND (or other DCC types). If "chat",
                       incoming data will be split in newline-separated
                       chunks. If "raw", incoming data is not touched.
        """
        with self.mutex:
            conn = DCCConnection(self, dcctype)
            self.connections.append(conn)
        return conn

    def _handle_event(self, connection, event):
        """
        Handle an Event event incoming on ServerConnection connection.
        """
        with self.mutex:
            matching_handlers = sorted(
                self.handlers.get("all_events", []) + self.handlers.get(event.type, [])
            )
            for handler in matching_handlers:
                result = handler.callback(connection, event)
                if result == "NO MORE":
                    return

    def _remove_connection(self, connection):
        """[Internal]"""
        with self.mutex:
            self.connections.remove(connection)
            self._on_disconnect(connection.socket)


_cmd_pat = (
    "^(@(?P<tags>[^ ]*) )?(:(?P<prefix>[^ ]+) +)?"
    "(?P<command>[^ ]+)( *(?P<argument> .+))?"
)
_rfc_1459_command_regexp = re.compile(_cmd_pat)


class DCCConnectionError(IRCError):
    pass


class DCCConnection(Connection):
    """
    A DCC (Direct Client Connection).

    DCCConnection objects are instantiated by calling the dcc
    method on a Reactor object.
    """

    socket = None
    connected = False
    passive = False
    peeraddress = None
    peerport = None

    def __init__(self, reactor, dcctype):
        super().__init__(reactor)
        self.dcctype = dcctype

    def connect(self, address, port):
        """Connect/reconnect to a DCC peer.

        Arguments:
            address -- Host/IP address of the peer.

            port -- The port number to connect to.

        Returns the DCCConnection object.
        """
        self.peeraddress = socket.gethostbyname(address)
        self.peerport = port
        self.buffer = buffer.LineBuffer()
        self.handlers = {}
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            self.socket.connect((self.peeraddress, self.peerport))
        except OSError as x:
            raise DCCConnectionError(f"Couldn't connect to socket: {x}") from x
        self.connected = True
        self.reactor._on_connect(self.socket)
        return self

    def listen(self, addr=None):
        """Wait for a connection/reconnection from a DCC peer.

        Returns the DCCConnection object.

        The local IP address and port are available as
        self.localaddress and self.localport.  After connection from a
        peer, the peer address and port are available as
        self.peeraddress and self.peerport.
        """
        self.buffer = buffer.LineBuffer()
        self.handlers = {}
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.passive = True
        default_addr = socket.gethostbyname(socket.gethostname()), 0
        try:
            self.socket.bind(addr or default_addr)
            self.localaddress, self.localport = self.socket.getsockname()
            self.socket.listen(10)
        except OSError as x:
            raise DCCConnectionError(f"Couldn't bind socket: {x}") from x
        return self

    def disconnect(self, message=""):
        """Hang up the connection and close the object.

        Arguments:

            message -- Quit message.
        """
        try:
            del self.connected
        except AttributeError:
            return

        try:
            self.socket.shutdown(socket.SHUT_WR)
        except OSError:
            pass

        self.socket.close()

        del self.socket
        self.reactor._handle_event(
            self, Event("dcc_disconnect", self.peeraddress, "", [message])
        )
        self.reactor._remove_connection(self)

    def process_data(self):
        """[Internal]"""

        if self.passive and not self.connected:
            conn, (self.peeraddress, self.peerport) = self.socket.accept()
            self.socket.close()
            self.socket = conn
            self.connected = True
            log.debug("DCC connection from %s:%d", self.peeraddress, self.peerport)
            self.reactor._handle_event(
                self, Event("dcc_connect", self.peeraddress, None, None)
            )
            return

        try:
            new_data = self.socket.recv(2**14)
        except OSError:
            # The server hung up.
            self.disconnect("Connection reset by peer")
            return
        if not new_data:
            # Read nothing: connection must be down.
            self.disconnect("Connection reset by peer")
            return

        if self.dcctype == "chat":
            self.buffer.feed(new_data)

            chunks = list(self.buffer)

            if len(self.buffer) > 2**14:
                # Bad peer! Naughty peer!
                log.info(
                    "Received >16k from a peer without a newline; " "disconnecting."
                )
                self.disconnect()
                return
        else:
            chunks = [new_data]

        command = "dccmsg"
        prefix = self.peeraddress
        target = None
        for chunk in chunks:
            log.debug("FROM PEER: %s", chunk)
            arguments = [chunk]
            log.debug(
                "command: %s, source: %s, target: %s, arguments: %s",
                command,
                prefix,
                target,
                arguments,
            )
            event = Event(command, prefix, target, arguments)
            self.reactor._handle_event(self, event)

    def privmsg(self, text):
        """
        Send text to DCC peer.

        The text will be padded with a newline if it's a DCC CHAT session.
        """
        if self.dcctype == 'chat':
            text += '\n'
        return self.send_bytes(self.encode(text))

    def send_bytes(self, bytes):
        """
        Send data to DCC peer.
        """
        try:
            self.socket.send(bytes)
            log.debug("TO PEER: %r\n", bytes)
        except OSError:
            self.disconnect("Connection reset by peer.")


class SimpleIRCClient:
    """A simple single-server IRC client class.

    This is an example of an object-oriented wrapper of the IRC
    framework.  A real IRC client can be made by subclassing this
    class and adding appropriate methods.

    The method on_join will be called when a "join" event is created
    (which is done when the server sends a JOIN messsage/command),
    on_privmsg will be called for "privmsg" events, and so on.  The
    handler methods get two arguments: the connection object (same as
    self.connection) and the event object.

    Functionally, any of the event names in `events.py` may be subscribed
    to by prefixing them with `on_`, and creating a function of that
    name in the child-class of `SimpleIRCClient`. When the event of
    `event_name` is received, the appropriately named method will be
    called (if it exists) by runtime class introspection.

    See `_dispatcher()`, which takes the event name, postpends it to
    `on_`, and then attemps to look up the class member function by
    name and call it.

    Instance attributes that can be used by sub classes:

        reactor -- The Reactor instance.

        connection -- The ServerConnection instance.

        dcc_connections -- A list of DCCConnection instances.
    """

    reactor_class = Reactor

    def __init__(self):
        self.reactor = self.reactor_class()
        self.connection = self.reactor.server()
        self.dcc_connections = []
        self.reactor.add_global_handler("all_events", self._dispatcher, -10)
        self.reactor.add_global_handler("dcc_disconnect", self._dcc_disconnect, -10)

    def _dispatcher(self, connection, event):
        """
        Dispatch events to on_<event.type> method, if present.
        """
        log.debug("_dispatcher: %s", event.type)

        def do_nothing(connection, event):
            return None

        method = getattr(self, "on_" + event.type, do_nothing)
        method(connection, event)

    def _dcc_disconnect(self, connection, event):
        self.dcc_connections.remove(connection)

    def connect(self, *args, **kwargs):
        """Connect using the underlying connection"""
        self.connection.connect(*args, **kwargs)

    def dcc(self, *args, **kwargs):
        """Create and associate a new DCCConnection object.

        Use the returned object to listen for or connect to
        a DCC peer.
        """
        dcc = self.reactor.dcc(*args, **kwargs)
        self.dcc_connections.append(dcc)
        return dcc

    def dcc_connect(self, address, port, dcctype="chat"):
        """Connect to a DCC peer.

        Arguments:

            address -- IP address of the peer.

            port -- Port to connect to.

        Returns a DCCConnection instance.
        """
        warnings.warn("Use self.dcc(type).connect()", DeprecationWarning, stacklevel=2)
        return self.dcc(dcctype).connect(address, port)

    def dcc_listen(self, dcctype="chat"):
        """Listen for connections from a DCC peer.

        Returns a DCCConnection instance.
        """
        warnings.warn("Use self.dcc(type).listen()", DeprecationWarning, stacklevel=2)
        return self.dcc(dcctype).listen()

    def start(self):
        """Start the IRC client."""
        self.reactor.process_forever()


class Event:
    """
    An IRC event.

    >>> print(Event('privmsg', '@somebody', '#channel'))
    type: privmsg, source: @somebody, target: #channel, arguments: [], tags: []
    """

    def __init__(self, type, source, target, arguments=None, tags=None):
        """
        Initialize an Event.

        Arguments:

            type -- A string describing the event.

            source -- The originator of the event (a nick mask or a server).

            target -- The target of the event (a nick or a channel).

            arguments -- Any event-specific arguments.
        """
        self.type = type
        self.source = source
        self.target = target
        if arguments is None:
            arguments = []
        self.arguments = arguments
        if tags is None:
            tags = []
        self.tags = tags

    def __str__(self):
        tmpl = (
            "type: {type}, "
            "source: {source}, "
            "target: {target}, "
            "arguments: {arguments}, "
            "tags: {tags}"
        )
        return tmpl.format(**vars(self))


def is_channel(string):
    """Check if a string is a channel name.

    Returns true if the argument is a channel name, otherwise false.
    """
    return string and string[0] in "#&+!"


def ip_numstr_to_quad(num):
    """
    Convert an IP number as an integer given in ASCII
    representation to an IP address string.

    >>> ip_numstr_to_quad('3232235521')
    '192.168.0.1'
    >>> ip_numstr_to_quad(3232235521)
    '192.168.0.1'
    """
    packed = struct.pack('>L', int(num))
    bytes = struct.unpack('BBBB', packed)
    return ".".join(map(str, bytes))


def ip_quad_to_numstr(quad):
    """
    Convert an IP address string (e.g. '192.168.0.1') to an IP
    number as a base-10 integer given in ASCII representation.

    >>> ip_quad_to_numstr('192.168.0.1')
    '3232235521'
    """
    bytes = map(int, quad.split("."))
    packed = struct.pack('BBBB', *bytes)
    return str(struct.unpack('>L', packed)[0])


class NickMask(str):
    """
    A nickmask (the source of an Event)

    >>> nm = NickMask('pinky!username@example.com')
    >>> nm.nick
    'pinky'

    >>> nm.host
    'example.com'

    >>> nm.user
    'username'

    >>> isinstance(nm, str)
    True

    >>> nm = NickMask('красный!red@yahoo.ru')

    >>> isinstance(nm.nick, str)
    True

    Some messages omit the userhost. In that case, None is returned.

    >>> nm = NickMask('irc.server.net')
    >>> nm.nick
    'irc.server.net'
    >>> nm.userhost
    >>> nm.host
    >>> nm.user
    """

    @classmethod
    def from_params(cls, nick, user, host):
        return cls('{nick}!{user}@{host}'.format(**vars()))

    @property
    def nick(self):
        nick, sep, userhost = self.partition("!")
        return nick

    @property
    def userhost(self):
        nick, sep, userhost = self.partition("!")
        return userhost or None

    @property
    def host(self):
        nick, sep, userhost = self.partition("!")
        user, sep, host = userhost.partition('@')
        return host or None

    @property
    def user(self):
        nick, sep, userhost = self.partition("!")
        user, sep, host = userhost.partition('@')
        return user or None

    @classmethod
    def from_group(cls, group):
        return cls(group) if group else None


def _ping_ponger(connection, event):
    "A global handler for the 'ping' event"
    connection.pong(event.target)



import re

_TAG_UNESCAPE_RE = re.compile(r'\\(.)')
_TAG_UNESCAPE_MAP = {':': ';', 's': ' ', 'n': '\n', 'r': '\r', '\\': '\\'}


class Tag:
    """
    An IRC message tag ircv3.net/specs/core/message-tags-3.2.html
    """

    @staticmethod
    def parse(item):
        r"""
        >>> Tag.parse('x') == {'key': 'x', 'value': None}
        True

        >>> Tag.parse('x=yes') == {'key': 'x', 'value': 'yes'}
        True

        >>> Tag.parse('x=3')['value']
        '3'

        >>> Tag.parse('x=red fox\\:green eggs')['value']
        'red fox;green eggs'

        >>> Tag.parse('x=red fox:green eggs')['value']
        'red fox:green eggs'

        >>> Tag.parse('x=a\\nb\\nc')['value']
        'a\nb\nc'

        >>> Tag.parse(r'x=a\bb\bc')['value']
        'abbbc'

        >>> Tag.parse(r'x=a\\nb\\nc')['value']
        'a\\nb\\nc'
        """
        key, sep, value = item.partition('=')
        value = _TAG_UNESCAPE_RE.sub(
            lambda char: _TAG_UNESCAPE_MAP.get(char.group(1), char.group(1)), value
        )
        value = value or None
        return {'key': key, 'value': value}

    @classmethod
    def from_group(cls, group):
        """
        Construct tags from the regex group
        """
        if not group:
            return
        tag_items = group.split(";")
        return list(map(cls.parse, tag_items))


class Arguments(list):
    @staticmethod
    def from_group(group):
        """
        Construct arguments from the regex group

        >>> Arguments.from_group('foo')
        ['foo']

        >>> Arguments.from_group(None)
        []

        >>> Arguments.from_group('')
        []

        >>> Arguments.from_group('foo bar')
        ['foo', 'bar']

        >>> Arguments.from_group('foo bar :baz')
        ['foo', 'bar', 'baz']

        >>> Arguments.from_group('foo bar :baz bing')
        ['foo', 'bar', 'baz bing']
        """
        if not group:
            return []

        main, sep, ext = group.partition(" :")
        arguments = main.split()
        if sep:
            arguments.append(ext)

        return arguments



#! /usr/bin/env python
#
# Example program using irc.client.
#
# This program is free without restrictions; do anything you like with
# it.
import argparse
import asyncio
import sys

import jaraco.logging

import irc.client
import irc.client_aio


class AioIRCCat(irc.client_aio.AioSimpleIRCClient):
    def __init__(self, target):
        irc.client.SimpleIRCClient.__init__(self)
        self.target = target

    def on_welcome(self, connection, event):
        if irc.client.is_channel(self.target):
            connection.join(self.target)
        else:
            self.future = asyncio.ensure_future(
                self.send_it(), loop=connection.reactor.loop
            )

    def on_join(self, connection, event):
        self.future = asyncio.ensure_future(
            self.send_it(), loop=connection.reactor.loop
        )

    def on_disconnect(self, connection, event):
        self.future.cancel()
        sys.exit(0)

    async def send_it(self):
        while 1:
            line = sys.stdin.readline().strip()
            if not line:
                break
            self.connection.privmsg(self.target, line)

            # Allow pause in the stdin loop to not block asyncio loop
            await asyncio.sleep(0)
        self.connection.quit("Using irc.client.py")


def get_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('server')
    parser.add_argument('nickname')
    parser.add_argument('target', help="a nickname or channel")
    parser.add_argument('--password', default=None, help="optional password")
    parser.add_argument('-p', '--port', default=6667, type=int)
    jaraco.logging.add_arguments(parser)
    return parser.parse_args()


def main():
    args = get_args()
    jaraco.logging.setup(args)
    target = args.target

    c = AioIRCCat(target)

    try:
        c.connect(args.server, args.port, args.nickname, password=args.password)
    except irc.client.ServerConnectionError as x:
        print(x)
        sys.exit(1)

    try:
        c.start()
    finally:
        c.connection.disconnect()
        c.reactor.loop.close()


if __name__ == "__main__":
    main()



"""
Handle Client-to-Client protocol per the `best available
spec <http://www.irchelp.org/irchelp/rfc/ctcpspec.html>`_.
"""

import re

LOW_LEVEL_QUOTE = '\x10'
LEVEL_QUOTE = "\\"
DELIMITER = '\x01'

low_level_mapping = {"0": '\x00', "n": "\n", "r": "\r", LEVEL_QUOTE: LEVEL_QUOTE}

low_level_regexp = re.compile(LOW_LEVEL_QUOTE + "(.)")


def _low_level_replace(match_obj):
    ch = match_obj.group(1)

    # If low_level_mapping doesn't have the character as key, we
    # should just return the character.
    return low_level_mapping.get(ch, ch)


def dequote(message):
    """
    Dequote a message according to CTCP specifications.

    The function returns a list where each element can be either a
    string (normal message) or a tuple of one or two strings (tagged
    messages).  If a tuple has only one element (ie is a singleton),
    that element is the tag; otherwise the tuple has two elements: the
    tag and the data.

    Arguments:

        message -- The message to be decoded.
    """

    # Perform the substitution
    message = low_level_regexp.sub(_low_level_replace, message)

    if DELIMITER not in message:
        return [message]

    # Split it into parts.
    chunks = message.split(DELIMITER)

    return list(_gen_messages(chunks))


def _gen_messages(chunks):
    i = 0
    while i < len(chunks) - 1:
        # Add message if it's non-empty.
        if len(chunks[i]) > 0:
            yield chunks[i]

        if i < len(chunks) - 2:
            # Aye!  CTCP tagged data ahead!
            yield tuple(chunks[i + 1].split(" ", 1))

        i = i + 2

    if len(chunks) % 2 == 0:
        # Hey, a lonely _CTCP_DELIMITER at the end!  This means
        # that the last chunk, including the delimiter, is a
        # normal message!  (This is according to the CTCP
        # specification.)
        yield DELIMITER + chunks[-1]



import socket


def identity(x):
    return x


class Factory:
    """
    A class for creating custom socket connections.

    To create a simple connection:

    .. code-block:: python

       server_address = ('localhost', 80)
       Factory()(server_address)

    To create an SSL connection:

    .. code-block:: python

       context = ssl.create_default_context()
       wrapper = functools.partial(context.wrap_socket, server_hostname=server_address)
       Factory(wrapper=wrapper)(server_address)

    To create an SSL connection with parameters to wrap_socket:

    .. code-block:: python

       context = ssl.create_default_context()
       wrapper = functools.partial(context.wrap_socket, server_hostname=server_address, ssl_cert=get_cert())
       Factory(wrapper=wrapper)(server_address)

    To create an IPv6 connection:

    .. code-block:: python

       Factory(ipv6=True)(server_address)

    Note that Factory doesn't save the state of the socket itself. The
    caller must do that, as necessary. As a result, the Factory may be
    re-used to create new connections with the same settings.

    """

    family = socket.AF_INET

    def __init__(self, bind_address=None, wrapper=identity, ipv6=False):
        self.bind_address = bind_address
        self.wrapper = wrapper
        if ipv6:
            self.family = socket.AF_INET6

    def connect(self, server_address):
        sock = self.wrapper(socket.socket(self.family, socket.SOCK_STREAM))
        self.bind_address and sock.bind(self.bind_address)
        sock.connect(server_address)
        return sock

    __call__ = connect


class AioFactory:
    """
    A class for creating async custom socket connections.

    To create a simple connection:

    .. code-block:: python

       server_address = ('localhost', 80)
       Factory()(protocol_instance, server_address)

    To create an SSL connection:

    .. code-block:: python

       Factory(ssl=True)(protocol_instance, server_address)

    To create an IPv6 connection:

    .. code-block:: python

       Factory(ipv6=True)(protocol_instance, server_address)

    Note that Factory doesn't save the state of the socket itself. The
    caller must do that, as necessary. As a result, the Factory may be
    re-used to create new connections with the same settings.

    """

    def __init__(self, **kwargs):
        self.connection_args = kwargs

    def connect(self, protocol_instance, server_address):
        return protocol_instance.loop.create_connection(
            lambda: protocol_instance, *server_address, **self.connection_args
        )

    __call__ = connect



from jaraco.collections import KeyTransformingDict

from . import strings


class IRCDict(KeyTransformingDict):
    """
    A dictionary of names whose keys are case-insensitive according to the
    IRC RFC rules.

    >>> d = IRCDict({'[This]': 'that'}, A='foo')

    The dict maintains the original case:

    >>> '[This]' in ''.join(d.keys())
    True

    But the keys can be referenced with a different case

    >>> d['a'] == 'foo'
    True

    >>> d['{this}'] == 'that'
    True

    >>> d['{THIS}'] == 'that'
    True

    >>> '{thiS]' in d
    True

    This should work for operations like delete and pop as well.

    >>> d.pop('A') == 'foo'
    True
    >>> del d['{This}']
    >>> len(d)
    0
    """

    @staticmethod
    def transform_key(key):
        if isinstance(key, str):
            key = strings.IRCFoldedCase(key)
        return key



"""
Internet Relay Chat (IRC) asyncio-based protocol client library.

This an extension of the IRC client framework in irc.client that replaces
original select-based event loop with one that uses Python 3's native
asyncio event loop.

This implementation shares many of the features of its select-based
cousin, including:

  * Abstraction of the IRC protocol.
  * Handles multiple simultaneous IRC server connections.
  * Handles server PONGing transparently.
  * Messages to the IRC server are done by calling methods on an IRC
    connection object.
  * Messages from an IRC server triggers events, which can be caught
    by event handlers.
  * Reading from and writing to IRC server sockets are normally done
    by an internal select() loop, but the select()ing may be done by
    an external main loop.
  * Functions can be registered to execute at specified times by the
    event-loop.
  * Decodes CTCP tagging correctly (hopefully); I haven't seen any
    other IRC client implementation that handles the CTCP
    specification subtleties.
  * A kind of simple, single-server, object-oriented IRC client class
    that dispatches events to instance methods is included.

Current limitations:
  * DCC chat has not yet been implemented
  * DCC file transfers are not suppored
  * RFCs 2810, 2811, 2812, and 2813 have not been considered.

Notes:
  * connection.quit() only sends QUIT to the server.
  * ERROR from the server triggers the error event and the disconnect event.
  * dropping of the connection triggers the disconnect event.
"""

import asyncio
import logging
import threading

from . import connection
from .client import (
    Event,
    Reactor,
    ServerConnection,
    ServerNotConnectedError,
    SimpleIRCClient,
    _ping_ponger,
)

log = logging.getLogger(__name__)


class IrcProtocol(asyncio.Protocol):
    """
    simple asyncio-based Protocol for handling connections to
    the IRC Server.

    Note: In order to maintain a consistent interface with
    `irc.ServerConnection`, handling of incoming and outgoing data
    is mostly handling by the `AioConnection` object, using the same
    callback methods as on an `irc.ServerConnection` instance.
    """

    def __init__(self, connection, loop):
        """
        Constructor for IrcProtocol objects.
        """
        self.connection = connection
        self.loop = loop

    def data_received(self, data):
        self.connection.process_data(data)

    def connection_lost(self, exc):
        log.debug(f"connection lost: {exc}")
        self.connection.disconnect()


class AioConnection(ServerConnection):
    """
    An IRC server connection.

    AioConnection objects are instantiated by calling the server
    method on a AioReactor object.

    Note: because AioConnection inherits from
    irc.client.ServerConnection, it has all the convenience
    methods on ServerConnection for handling outgoing data,
    including (but not limited to):

        * join(channel, key="")
        * part(channel, message="")
        * privmsg(target, text)
        * privmsg_many(targets, text)
        * quit(message="")

    And many more.  See the documentation on
    irc.client.ServerConnection for a full list of convience
    functions available.
    """

    protocol_class = IrcProtocol

    async def connect(
        self,
        server,
        port,
        nickname,
        password=None,
        username=None,
        ircname=None,
        connect_factory=connection.AioFactory(),
    ):
        """Connect/reconnect to a server.

        Arguments:

        * server - Server name
        * port - Port number
        * nickname - The nickname
        * password - Password (if any)
        * username - The username
        * ircname - The IRC name ("realname")

        * connect_factory - An async callable that takes the event loop and the
          server address, and returns a connection (with a socket interface)

        This function can be called to reconnect a closed connection.

        Returns the AioProtocol instance (used for handling incoming
        IRC data) and the transport instance (used for handling
        outgoing data).
        """
        if self.connected:
            self.disconnect("Changing servers")

        self.buffer = self.buffer_class()
        self.handlers = {}
        self.real_server_name = ""
        self.real_nickname = nickname
        self.server = server
        self.port = port
        self.server_address = (server, port)
        self.nickname = nickname
        self.username = username or nickname
        self.ircname = ircname or nickname
        self.password = password
        self.connect_factory = connect_factory

        protocol_instance = self.protocol_class(self, self.reactor.loop)
        connection = self.connect_factory(protocol_instance, self.server_address)
        transport, protocol = await connection

        self.transport = transport
        self.protocol = protocol

        self.connected = True
        self.reactor._on_connect(self.protocol, self.transport)

        # Log on...
        if self.password:
            self.pass_(self.password)
        self.nick(self.nickname)
        self.user(self.username, self.ircname)
        return self

    def process_data(self, new_data):
        """
        handles incoming data from the `IrcProtocol` connection.
        Main data processing/routing is handled by the _process_line
        method, inherited from `ServerConnection`
        """
        self.buffer.feed(new_data)

        # process each non-empty line after logging all lines
        for line in self.buffer:
            log.debug("FROM SERVER: %s", line)
            if not line:
                continue
            self._process_line(line)

    def send_raw(self, string):
        """Send raw string to the server, via the asyncio transport.

        The string will be padded with appropriate CR LF.
        """
        log.debug(f'RAW: {string}')
        if self.transport is None:
            raise ServerNotConnectedError("Not connected.")

        self.transport.write(self._prep_message(string))

    def disconnect(self, message=""):
        """Hang up the connection.

        Arguments:
            message -- Quit message.
        """
        try:
            del self.connected
        except AttributeError:
            return

        self.quit(message)

        self.transport.close()

        self._handle_event(Event("disconnect", self.server, "", [message]))


class AioReactor(Reactor):
    """
    Processes message from on or more asyncio-based IRC server connections.

    This class inherits most of its functionality from irc.client.Reactor,
    and mainly replaces the select-based loop from that reactor with
    an asyncio event loop.

    Note: if not event loop is passed into AioReactor, it will try to get
    the current loop.  However, if there is a specific event loop
    you'd like to use, you can pass that loop by using the `loop` kwarg
    in the AioReactor's constructor:

        async def my_repeating_message(connection):
            while True:
                connection.privmsg('#my-irc-channel', 'hello!')
                await asyncio.sleep(60)

        my_loop = asyncio.new_event_loop()

        client = AioReactor(loop=my_loop)
        server = client.server()

        # use `my_loop` to initialize the repeating message on the same
        # loop as the AioRector
        asyncio.ensure_future(my_repeating_message(server), loop=my_loop)

        # connect to the server, start the loop
        server.connect('my.irc.server', 6667, 'my_irc_nick')
        client.process_forever()

    The above code will connect to the IRC server my.irc.server
    and echo the 'hello!' message to 'my-irc-channel' every 60 seconds.
    """

    connection_class = AioConnection

    def __do_nothing(*args, **kwargs):
        pass

    def __init__(self, on_connect=__do_nothing, on_disconnect=__do_nothing, loop=None):
        self._on_connect = on_connect
        self._on_disconnect = on_disconnect

        self.connections = []
        self.handlers = {}

        self.mutex = threading.RLock()

        self.loop = asyncio.new_event_loop() if loop is None else loop

        self.add_global_handler("ping", _ping_ponger, -42)

    def process_forever(self):
        """Run an infinite loop, processing data from connections.

        Rather than call `process_once` repeatedly, like
        irc.client.reactor, incoming data is handled via the
        asycio.Protocol class -- by default, the IrcProtocol
        class definied above.
        """
        self.loop.run_forever()


class AioSimpleIRCClient(SimpleIRCClient):
    """A simple single-server IRC client class.

    This class is functionally equivalent irc.client.SimpleIRCClient,
    the only difference being the using of the AioReactor for
    asyncio-based loops.

    For more information on AioSimpleIRCClient, see the documentation
    on irc.client.SimpleIRCClient
    """

    reactor_class = AioReactor

    def connect(self, *args, **kwargs):
        self.reactor.loop.run_until_complete(self.connection.connect(*args, **kwargs))
